#include "api/calibrate_controller.h"

#include <limits>

#include <esp_log.h>
#include <kbf/driver/ultrasonic.h>
#include <kbf.h>

#include "settings.h"
#include "tasks/display_manager.h"
#include "tasks/ultrasonic.h"

using namespace kbf;
using http::Request;
using http::Response;

Response CalibrateController::get(const Request &request) {
    ESP_LOGD(TAG, "%s()", __func__);

    auto config = dk::settings::calibration();
    return Response({{"high", config.high},
                     {"low",  config.low}});
}

Response CalibrateController::post(const Request &request) {
    ESP_LOGD(TAG, "%s()", __func__);
    DisplayManager::setMessage("calibrating...");

    Ultrasonic::disable();
    auto sensor = driver::Ultrasonic(CONFIG_DK_ULTRASONIC_TRIGGER, CONFIG_DK_ULTRASONIC_ECHO);

    int  min   = std::numeric_limits<int>::max();
    int  max   = 0;
    bool error = false;

    for (int i = 0; i < CONFIG_DK_ULTRASONIC_CALIB_SAMPLE_SIZE; i++) {
        auto reading = (int) sensor.read();
        if (reading >= 0x200) {
            ESP_LOGW(TAG, "reading failed: %d", reading);
            error = true;
            break;
        }

        ESP_LOGD(TAG, "  reading: %d", reading);
        if (reading <= min) {
            min = reading - 1;
        } else if (reading >= max) {
            max = reading + 1;
        }

        kbf::sleep(CONFIG_DK_ULTRASONIC_CALIB_SLEEP);
    }

    ESP_LOGI(TAG, "%s: min = %d, max = %d", __func__, min, max);

    dk::settings::Calibration calib = {.high = max, .low = min};
    if (error) {
        DisplayManager::setMessage("cal. error", DisplayManager::defaultTimeout);
        DisplayManager::setDoorState(DisplayManager::DoorState::NOT_CALIBRATED);
        calib = {.high = 0, .low = 0};
        dk::settings::setCalibration(calib);
    } else {
        DisplayManager::setMessage("cal: " + std::to_string(calib.low) + " - " + std::to_string(calib.high),
                                   DisplayManager::defaultTimeout);
        dk::settings::setCalibration(calib);
        Ultrasonic::enable();
    }

    return Response({
                            {"high", calib.high},
                            {"low",  calib.low}
                    }, error ? 500 : 200);
}
