#include "tasks/boot.h"

#include <memory>

#include <esp_log.h>

#include <kbf/wifi.h>

#include "dk.h"
#include "api/calibrate_controller.h"
#include "api/ping_controller.h"
#include "api/unpair_controller.h"
#include "api/pairing_id_middleware.h"
#include "pairing_service/pair_controller.h"
#include "pairing_service/settings_controller.h"
#include "pairing_service/verify_controller.h"
#include "tasks/display_manager.h"
#include "tasks/event_manager.h"
#include "tasks/main.h"
#include "tasks/ultrasonic.h"

using namespace kbf;
using namespace dk;
using std::make_shared;

extern const unsigned char cert_start[] asm("_binary_server_pem_start");
extern const unsigned char cert_end[]   asm("_binary_server_pem_end");

Boot::Boot(const std::string &name, void *arg, uint32_t stackSize, uint32_t priority) :
        Task(name, arg, stackSize, priority) {}

void Boot::run(void *arg) {
    ESP_LOGI(TAG, "%s", __func__);

#ifdef CONFIG_DK_PRESET_PAIRING_ID_ENABLED
    ESP_LOGW(TAG, "using preset pairing ID \"%s\"", CONFIG_DK_PRESET_PAIRING_ID);
    dk::settings::setPairingId(CONFIG_DK_PRESET_PAIRING_ID);
#endif

    ESP_LOGI(TAG, "starting DisplayManager");
    DisplayManager::start();
    DisplayManager::setStatus("DoorKeeper");
    DisplayManager::setMessage("loading...");

    ESP_LOGI(TAG, "setting up mainEventGroup");
    mainEventGroup = make_shared<rtos::EventGroup>();

    ESP_LOGI(TAG, "setting up button driver");
    button = make_shared<driver::Button>(CONFIG_DK_BUTTON1);
    button->onRelease   = {[](driver::Button &) {
        ESP_LOGI(TAG, "button->onRelease()");
        mainEventGroup->setBit(MainEvents::SEND_TEST_MESSAGE);
    }};
    button->onLongPress = {[](driver::Button &) {
        ESP_LOGI(TAG, "button->onLongPress()");
        clearPairing();
    }};

    ESP_LOGI(TAG, "starting networking");
    wifi::sta::onIp         = ([]() { mainEventGroup->setBit(MainEvents::SET_ONLINE_MODE); });
    wifi::sta::onDisconnect = ([]() { mainEventGroup->setBit(MainEvents::SET_OFFLINE_MODE); });
    wifi::sta::start();
    http::Client::addCert(cert_start, cert_end);

    ESP_LOGI(TAG, "setting up pairingService");
    pairingService = make_shared<WebService>();
    pairingService->controller<PairController>();
    pairingService->controller<SettingsController>();
    pairingService->controller<VerifyController>();

    ESP_LOGI(TAG, "setting up apiService");
    apiService = make_shared<WebService>();
    apiService->controller<CalibrateController>();
    apiService->controller<PingController>();
    apiService->controller<UnpairController>();
    apiService->middleware<PairingIdMiddleware>();

    ESP_LOGI(TAG, "starting EventManager");
    EventManager::start();

    ESP_LOGI(TAG, "starting Ultrasonic");
    Ultrasonic::start();

    ESP_LOGI(TAG, "starting Main");
    task::start<Main>("main", nullptr, 2 << 11);

    ESP_LOGI(TAG, "setting mode");
    if (paired()) {
        auto wifiConfig = dk::settings::wifi();

        if (wifiConfig.ssid.empty() || wifiConfig.password.empty()) {
            ESP_LOGE(TAG, "paired without wifi config, erasing NVS and restarting");
            kbf::nvs::erase();
            esp_restart();
        }

        DisplayManager::setWifiState(DisplayManager::WifiState::CONNECTING);
        DisplayManager::setMessage("connecting...");
        wifi::sta::connect(wifiConfig.ssid, wifiConfig.password);
        EventManager::event(EventManager::Event::BOOT);
        Ultrasonic::enable();
    } else {
        mainEventGroup->setBit(MainEvents::SET_PAIRING_MODE);
    }

    ESP_LOGI(TAG, "finished");
}
