#include "tasks/event_manager.h"

#include <string>
#include <memory>

#include <esp_log.h>
#include <kbf.h>
#include <kbf/http/exception.h>

#include "dk.h"
#include "settings.h"
#include "tasks/display_manager.h"

using namespace kbf;
using std::nullopt;
using std::to_string;
using std::shared_ptr;

Queue<EventManager::Event> EventManager::eventQueue{CONFIG_DK_EVENT_QUEUE_SIZE};

static shared_ptr<Task> instance;

[[noreturn]] void EventManager::run(void *arg) {
    ESP_LOGI(TAG, "%s()", __func__);

    pairingId = dk::settings::pairingId();
    ESP_LOGI(TAG, "pairingId = \"%s\"", pairingId.c_str());

    while (true) {
        lastEvent = eventQueue.pop(noopInterval);

        if (lastEvent == nullopt) {
            postEvent(Event::NOOP);
            continue;
        }

        while (!postEvent(lastEvent.value())) {
            ESP_LOGI(TAG, "retrying in %d ms...", CONFIG_DK_EVENT_HTTP_RETRY_INTERVAL);
            kbf::sleep(CONFIG_DK_EVENT_HTTP_RETRY_INTERVAL);
        }
    }
}

bool EventManager::event(EventManager::Event event) {
    ESP_LOGI(TAG, "%s(%d); queue size = %d", __func__, (int) event, eventQueue.size());

    switch (event) {
       case Event::DOOR_OPEN:
            DisplayManager::setDoorState(DisplayManager::DoorState::OPEN);
            break;
        case Event::DOOR_CLOSED:
            DisplayManager::setDoorState(DisplayManager::DoorState::CLOSED);
            break;
        case Event::BOOT:
            break;
        case Event::NOOP:
            break;
    }

    auto result = eventQueue.push(event);
    if (!result) ESP_LOGW(TAG, "%s: event queue is full", __func__);

    return result;
}

bool EventManager::postEvent(Event event) {
    ESP_LOGI(TAG, "%s(%d)", __func__, (int) event);

    http::Client client(CONFIG_DK_HTTP_CLIENT_TIMEOUT);  // TODO reuse client

    try {
        auto response = client.post(API_URL, {{"type", event}}, {{{"X-Pairing-ID", pairingId}}});

        if (!response) {
            ESP_LOGE(TAG, "%s: request failed with unknown error", __func__);
            return false;
        } else if (response->status == 403) {
            ESP_LOGE(TAG, "%s: received 403, resetting pairing", __func__);
            dk::clearPairing();
            return false;
        } else if (response->status == 200) {
            noopInterval = response->json().find("noop_interval")->get<int>() * 1000;
            ESP_LOGI(TAG, "%s: success; id = %s; noop_interval = %d", __func__,
                     response->json().find("id")->get<std::string>().c_str(), noopInterval);
            return true;
        }

        ESP_LOGE(TAG, "%s: response.status = %d", __func__, response->status);
        return false;

    } catch (http::exception::ConnectionError &e) {
        ESP_LOGW(TAG, "%s: connection refused", __func__);
        return false;
    } catch (http::exception::Timeout &e) {
        ESP_LOGW(TAG, "%s: request timeout", __func__);
        return false;
    }
}

void EventManager::start() {
    ESP_LOGI(TAG, "%s", __func__);

    if (instance && instance->running()) {
        ESP_LOGE(TAG, "%s: already running", __func__);
        return;
    }

    if (dk::settings::pairingId().empty()) {
        ESP_LOGW(TAG, "%s: no pairing ID, aborting", __func__);
        return;
    }

    instance = task::start<EventManager>("event_manager", nullptr, 2 << 14);
}

void EventManager::stop() {
    ESP_LOGI(TAG, "%s", __func__);

    if (!instance || !instance->running()) {
        ESP_LOGE(TAG, "%s: not running", __func__);
        return;
    }

    instance->stop();
    instance = nullptr;
}

