#include "tasks/main.h"

#include <esp_log.h>

#include <kbf/wifi.h>
using namespace kbf;

#include "tasks/display_manager.h"
#include "tasks/event_manager.h"

#include "dk.h"
using namespace dk;

static net::MDNS *mdns;

extern const unsigned char cert_start[] asm("_binary_cert_pem_start");
extern const unsigned char cert_end[]   asm("_binary_cert_pem_end");
extern const unsigned char key_start[] asm("_binary_key_pem_start");
extern const unsigned char key_end[]   asm("_binary_key_pem_end");

Main::Main(const std::string &name, void *arg, uint32_t stackSize, uint32_t priority) :
        Task(name, arg, stackSize, priority) {}

string generatePin() {
#ifdef CONFIG_DK_PRESET_PIN_ENABLED
    static_assert(std::char_traits<char>::length(CONFIG_DK_PRESET_PIN) >= 8, "pairing PIN must be at least 8 chars");
    return CONFIG_DK_PRESET_PIN;
#else
    string pin;
    for (int i = 0; i < 8; i++) pin += std::to_string(esp_random() % 10);
    return pin;
#endif
}

void setupPairingMode() {
    ESP_LOGI(Main::TAG, "%s()", __func__);
    DisplayManager::setStatus("pairing");

    string pin = generatePin();
    ESP_LOGI(Main::TAG, "pairing PIN: \"%s\"", pin.c_str());
    DisplayManager::setMessage("PIN: " + pin);

    apiService->stop();
    pairingService->stop();
    wifi::stop();

    wifi::ap::onConnect = {[](wifi::STA &){ DisplayManager::setMessage("connected"); }};
    wifi::ap::start("doorkeeper", pin);
    DisplayManager::setWifiState(DisplayManager::WifiState::AP);
    pairingService->startSSL(cert_start, cert_end, key_start, key_end);
}

void startMdns() {
    ESP_LOGI(Main::TAG, "%s", __func__);
    delete mdns;
    mdns = new net::MDNS("doorkeeper");
    mdns->addService("_http", 80, "DoorKeeper API");
}

void setupOnlineMode() {
    ESP_LOGI(Main::TAG, "%s()", __func__);
    DisplayManager::setStatus("online");
    DisplayManager::setMessage("");
    DisplayManager::setWifiState(DisplayManager::WifiState::ONLINE);

    pairingService->stop();
    apiService->startSSL(cert_start, cert_end, key_start, key_end);
    startMdns();
}

void stopAP() {
    ESP_LOGI(Main::TAG, "%s()", __func__);

    pairingService->stop();
    wifi::ap::stop();
    DisplayManager::setWifiState(DisplayManager::WifiState::ONLINE);

    pairingService->startSSL(cert_start, cert_end, key_start, key_end);
    startMdns();
}

[[noreturn]] void Main::run(void *arg) {
    ESP_LOGI(TAG, "%s", __func__);

    bool testDoorOpen = false;

    while (true) {
        mainEventGroup->waitForAny(ALL_EVENTS);

        if (mainEventGroup->getBit(MainEvents::STOP_AP)) {
            ESP_LOGI(TAG, "event: STOP_AP");
            stopAP();
            mainEventGroup->clearBit(MainEvents::STOP_AP);
        }

        if (mainEventGroup->getBit(MainEvents::SET_PAIRING_MODE)) {
            ESP_LOGI(TAG, "event: SET_PAIRING_MODE");
            setupPairingMode();
            mainEventGroup->clearBit(MainEvents::SET_PAIRING_MODE);
        }

        if (mainEventGroup->getBit(MainEvents::SET_ONLINE_MODE)) {
            ESP_LOGI(TAG, "event: SET_ONLINE_MODE");
            setupOnlineMode();
            mainEventGroup->clearBit(MainEvents::SET_ONLINE_MODE);
        }

        if (mainEventGroup->getBit(MainEvents::SET_OFFLINE_MODE)) {
            ESP_LOGI(TAG, "event: SET_OFFLINE_MODE");
            // TODO reconnect
            mainEventGroup->clearBit(MainEvents::SET_OFFLINE_MODE);
        }

        if (mainEventGroup->getBit(MainEvents::SEND_TEST_MESSAGE)) {
            ESP_LOGI(TAG, "event: SEND_TEST_MESSAGE");
//            testNotification();
            testDoorOpen = !testDoorOpen;
            if (testDoorOpen) EventManager::event(EventManager::Event::DOOR_OPEN);
            else EventManager::event(EventManager::Event::DOOR_CLOSED);
            mainEventGroup->clearBit(MainEvents::SEND_TEST_MESSAGE);
        }
    }
}
