#include "tasks/ultrasonic.h"

#include <atomic>

#include <esp_log.h>

#include <kbf.h>
#include <kbf/driver/ultrasonic.h>

#include "settings.h"
#include "tasks/event_manager.h"

using namespace kbf;
using std::shared_ptr;
using std::make_shared;
using std::atomic;

static const int BIT_ENABLED = 0;

shared_ptr <rtos::EventGroup> Ultrasonic::eventGroup;

atomic<int>  low{0};
atomic<int>  high{0};
atomic<bool> doorOpen{false};
static int   confidence;

inline bool isOpen(uint32_t reading) { return reading < low || reading > high; }

[[noreturn]] void Ultrasonic::run(void *arg) {
    eventGroup = make_shared<rtos::EventGroup>();
    auto sensor = driver::Ultrasonic(CONFIG_DK_ULTRASONIC_TRIGGER, CONFIG_DK_ULTRASONIC_ECHO);

    confidence = 0;
    bool first = true;

    while (true) {
        eventGroup->waitForBit(BIT_ENABLED);

        auto reading = sensor.read();
        ESP_LOGV(TAG, "reading: %d; h/l: %d/%d", reading, high.load(), low.load());

        if (first) {
            /** trigger confidence check on first run */
            doorOpen = isOpen(reading) ? false : true;
            first = false;
        }

        if (isOpen(reading)) {
            if (confidence > 0) confidence = 0;
            if (!doorOpen) {
                ESP_LOGI(TAG, "open confidence %d/%d", confidence - 1, OPEN_CONFIDENCE);
                if (--confidence == OPEN_CONFIDENCE) {
                    doorOpen = true;
                    EventManager::event(EventManager::Event::DOOR_OPEN);
                }
            }
        } else {
            if (confidence < 0) confidence = 0;
            if (doorOpen) {
                ESP_LOGI(TAG, "closed confidence %d/%d", confidence + 1, CLOSED_CONFIDENCE);
                if (++confidence == CLOSED_CONFIDENCE) {
                    doorOpen = false;
                    EventManager::event(EventManager::Event::DOOR_CLOSED);
                }
            }
        }

        kbf::sleep(READING_INTERVAL_MS);
    }
}

void Ultrasonic::start() {
    ESP_LOGI(TAG, "%s", __func__);
    task::start<Ultrasonic>("ultrasonic");
}

bool Ultrasonic::enable() {
    ESP_LOGI(TAG, "%s", __func__);

    auto calibration = dk::settings::calibration();
    if (!calibration.low || !calibration.high) {
        ESP_LOGW(TAG, "not calibrated");
        return false;
    }

    if (dk::settings::pairingId().empty()) {
        ESP_LOGW(TAG, "not paired");
        return false;
    }

    low  = calibration.low;
    high = calibration.high;
    eventGroup->setBit(BIT_ENABLED);

    return true;
}

void Ultrasonic::disable() {
    ESP_LOGI(TAG, "%s", __func__);

    if (!eventGroup->getBit(BIT_ENABLED)) {
        ESP_LOGW(TAG, "already disabled");
        return;
    }

    eventGroup->clearBit(BIT_ENABLED);
}
